/*
  sdlhack - Force SDL games to minimize when Ctrl-Z is pressed

  Copyright (C) 2010 Katie Stafford (katie@ktpanda.org)

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <SDL.h>

/* dlopen() handle for libSDL */
static void* sdlhandle;

/* Function pointers to real SDL library. */
static int (*orig_SDL_Init)(Uint32 mask);
static int (*orig_SDL_WasInit)(Uint32 mask);
static void (*orig_SDL_SetError)(const char* fmt, ...);
static SDL_Surface* (*orig_SDL_SetVideoMode)(int width, int height, int bpp, Uint32 flags);
static void (*orig_SDL_SetEventFilter)(SDL_EventFilter filter);
static SDL_EventFilter (*orig_SDL_GetEventFilter)();
static void (*orig_SDL_WM_IconifyWindow)(); 

/* These are only defined if joystick support is not disabled */
static int (*orig_SDL_NumJoysticks)(void);
static SDL_Joystick *(*orig_SDL_JoystickOpen)(int device_index);
static const char *(*orig_SDL_JoystickName)(int device_index);
static int (*orig_SDL_JoystickOpened)(int device_index);
static int (*orig_SDL_JoystickEventState)(int state);

/* Event filter set by the application, if any. */
static SDL_EventFilter app_event_filter;

static SDLCALL int our_event_filter(const SDL_Event* evt);

static void* load_sym(const char* name) {
    void* sym = dlsym(sdlhandle, name);
    if (!sym) {
        printf("failed loading %s, err = %s\n", name, dlerror());
        exit(1);
    }
    printf ("  %-22s = %p\n", name, sym);
    return sym;
}

static void load_sdl_funcs() {
    char* joydisable;

    sdlhandle = dlopen("libSDL-1.2.so.0", RTLD_NOW);
    if (!sdlhandle) {
        fprintf(stderr, "Could not load SDL: %s\n", dlerror());
        /* What to do? The game called SDL_Init, but SDL can't load?
           Probably best to just bail. */
        exit(1);
    }
    printf ("SDL iconify hack initializing, dlopen handle = %p\n", sdlhandle);

#define SYM_INIT(sym) orig_##sym = load_sym(#sym);

    SYM_INIT(SDL_Init);
    SYM_INIT(SDL_WasInit);
    SYM_INIT(SDL_SetError);

    SYM_INIT(SDL_SetVideoMode);
    SYM_INIT(SDL_SetEventFilter);
    SYM_INIT(SDL_GetEventFilter);
    SYM_INIT(SDL_WM_IconifyWindow);

    if ((joydisable = getenv("SDL_DISABLE_JOYSTICK")) == NULL ||
        atoi(joydisable) == 0) {
        
        SYM_INIT(SDL_NumJoysticks);
        SYM_INIT(SDL_JoystickOpen);
        SYM_INIT(SDL_JoystickName);
        SYM_INIT(SDL_JoystickOpened);
        SYM_INIT(SDL_JoystickEventState);
    } else {
        printf("joystick disabled\n");
    }

#undef SYM_INIT

}

static void check_event_filter() {
    if (orig_SDL_WasInit(SDL_INIT_VIDEO) && orig_SDL_GetEventFilter() != our_event_filter)
        orig_SDL_SetEventFilter(our_event_filter);
}

int SDL_Init(Uint32 mask) {
    int rc;
    if (!sdlhandle)
        load_sdl_funcs();
 
    if (!orig_SDL_NumJoysticks)
        mask &= ~SDL_INIT_JOYSTICK;

    rc = orig_SDL_Init(mask);

    if (rc == 0)
        check_event_filter();

    return rc;
}

static SDLCALL int our_event_filter(const SDL_Event* evt) {
    /* Check for Ctrl-Z */
    if (evt->type == SDL_KEYDOWN && 
        evt->key.keysym.mod & (KMOD_LCTRL|KMOD_RCTRL) && 
        evt->key.keysym.sym == SDLK_z) {

        printf("Iconify hack: Iconifying window\n");
        orig_SDL_WM_IconifyWindow();

        /* eat the event */
        return 0;
    }

    /* call application event filter */
    if (app_event_filter) return (*app_event_filter)(evt);
    return 1;
}

/* Recheck event filter after SDL_SetVideoMode*/
SDL_Surface* SDL_SetVideoMode(int width, int height, int bpp, Uint32 flags) {
    SDL_Surface* ret;

    if (!sdlhandle)
        load_sdl_funcs();
 
    ret = orig_SDL_SetVideoMode(width, height, bpp, flags);
    if (ret)
        check_event_filter();

    return ret;
}

void SDL_SetEventFilter(SDL_EventFilter filter) {
    app_event_filter = filter;
    check_event_filter();
}

SDL_EventFilter SDL_GetEventFilter() {
    return app_event_filter;
}

int SDL_NumJoysticks(void)
{

    if (orig_SDL_NumJoysticks)
        return orig_SDL_NumJoysticks();

    return 0;
}

SDL_Joystick *SDL_JoystickOpen(int device_index)
{
    if (orig_SDL_JoystickOpen)
        return orig_SDL_JoystickOpen(device_index);

    if (orig_SDL_SetError)
        orig_SDL_SetError("Joystick disabled");
    return NULL;
}

const char *SDL_JoystickName(int device_index)
{
    if (orig_SDL_JoystickName)
        return orig_SDL_JoystickName(device_index);
    if (orig_SDL_SetError)
        orig_SDL_SetError("Joystick disabled");
    return NULL;
}

int SDL_JoystickOpened(int device_index)
{
    if (orig_SDL_JoystickOpened)
        return orig_SDL_JoystickOpened(device_index);
    return 0;
}

/*
  Emulate the behavior of the original, but don't do anything.
 */

int SDL_JoystickEventState(int state)
{
    static int cstate = SDL_IGNORE;

    if (orig_SDL_JoystickEventState)
        return orig_SDL_JoystickEventState(state);

    if (state == SDL_QUERY)
        return cstate;
    cstate = state;
    return cstate;
}

